"""
AWS Secrets Manager
"""
import datetime
import uuid
from decimal import Decimal
from pathlib import Path
import marshmallow as ma

from .util import AttrDict
from .parser import SecretParser
from . import fields


class NotSet:
    """
    NotSet is a placeholder for when a value is not set
    """
    pass


class SecretManager(SecretParser):
    """
    Secret Manager resolves secrets from AWS Secrets Manager
    This inherits from SecretParser that handles connecting to AWS and parsing the secret(s)

    Methods for type casting values:
    * `str`: String - This is the default type
    * `int`: Integer
    * `float`: Float
    * `decimal`: decimal.Decimal
    * `bool`: Boolean
    * `list`: List - You can specify a `delimiter` and a `subcast` type for list elements
    * `choices`: List[tuple] - You can specify a `delimiter` and a `subcast` type for list elements.
        Returns the form: [(key, value), (key, value)]
    * `datetime`: datetime.datetime - You can specify a `format` for the datetime string
    * `date`: datetime.date - You can specify a `format` for the date string
    * `time`: datetime.time - You can specify a `format` for the time string
    * `timedelta`: datetime.timedelta - You can specify a `format` for the timedelta string
    * `timedelta_seconds`: datetime.timedelta - Parses from seconds
    * `uuid`: uuid.UUID - You can specify a `version` for the UUID. 1=Time-based, 3=Name-based, 4=Random, 5=Name-based
    * `log_level`: int - Parses a log level string to an int
    * `path`: pathlib.Path - Parses a string to a pathlib.Path object
    * `dict`: AttrDict - You can specify a `prefix` for the dictionary keys, a `subcast_keys` type,
        and a `subcast_values` type
    """

    def value(self, name, default=NotSet) -> str:
        """
        Get the value of a secret.  Returns raw format.
        """
        if not self._secrets:
            self.load()
        # Try to get the value from secrets
        for secret in reversed(self._secrets.values()):  # We want to get the most recent secret first
            try:
                return secret.SecretValues[name]
            except KeyError:
                pass
        # Try to get value from environment variables
        if name in self.env:
            return self.env[name]
        elif name.upper() in self.env:
            return self.env[name.upper()]

        if default is not NotSet:
            return default

        # No value found
        raise KeyError(f'Key "{name}" is not found.')

    def str(self, name, default: (str, NotSet) = NotSet) -> str:
        """
        Get the value of a secret as a string
        """
        return fields.Str().deserialize(self.value(name, default=default))

    def int(self, name, default: (str, NotSet) = NotSet) -> int:
        """
        Get the value of a secret as an integer
        """
        return fields.Int().deserialize(self.value(name, default=default))

    def float(self, name, default: (str, NotSet) = NotSet) -> float:
        """
        Get the value of a secret as a float
        """
        return fields.Float().deserialize(self.value(name, default=default))

    def decimal(self, name, default: (str, NotSet) = NotSet) -> Decimal:
        """
        Get the value of a secret as a decimal
        """
        return fields.Decimal().deserialize(self.value(name, default=default))

    def bool(self, name, default: (str, NotSet) = NotSet) -> bool:
        """
        Get the value of a secret as a boolean
        """
        return fields.Bool().deserialize(self.value(name, default=default))

    def list(self, name, delimiter=',', subcast=fields.Str, default: (str, NotSet) = NotSet) -> list:
        """
        Get the value of a secret as a list
        """
        if isinstance(subcast, type):
            subcast = subcast()
        return [subcast.deserialize(v) for v in self.value(name, default=default).split(delimiter)]

    def choices(self, name, delimiter=',', subcast: ma.fields.Field = fields.Str,
                default: (str, NotSet) = NotSet) -> list:
        """
        Get the value of a secret as a list of tuples
        """
        return fields.Choices(delimiter=delimiter, subcast=subcast).deserialize(self.value(name, default=default))

    def datetime(self, name, format='%Y-%m-%d %H:%M:%S', default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a datetime
        """
        return fields.Datetime(format=format).deserialize(self.value(name, default=default))

    def date(self, name, format='%Y-%m-%d', default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a date
        """
        return fields.Date(format=format).deserialize(self.value(name, default=default))

    def time(self, name, format='%H:%M:%S', default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a time
        """
        return fields.Time(format=format).deserialize(self.value(name, default=default))

    def timedelta(self, name, default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a timedelta
        Format: HH:MM:SS
        """
        value = self.value(name, default=default)
        return datetime.timedelta(**{k: int(v) for k, v in zip(['hours', 'minutes', 'seconds'], value.split(':'))})

    def timedelta_seconds(self, name, default: (str, NotSet) = NotSet) -> datetime:
        """
        Get the value of a secret as a timedelta
        """
        return datetime.timedelta(seconds=self.int(name, default=default))

    def uuid(self, name, version=4, default: (str, NotSet) = NotSet) -> uuid:
        """
        Get the value of a secret as a UUID
        """
        return fields.UUID(metadata=dict(version=version)).deserialize(self.value(name, default=default))

    def log_level(self, name, default: (str, NotSet) = NotSet) -> int:
        """
        Get the value of a secret as a log level
        """
        return fields.LogLevel().deserialize(self.value(name, default=default))

    def path(self, name, default: (str, NotSet) = NotSet) -> Path:
        """
        Get the value of a secret as a Path
        """
        return fields.Path().deserialize(self.value(name, default=default))

    def _parse_dict(self, prefix, dictionary: dict, response_dict: AttrDict = None,
                    subcast_keys: ma.fields.Field = fields.Str,
                    subcast_values: ma.fields.Field = fields.Str) -> AttrDict:
        """
        Parse a dictionary from a dictionary of any type.

        WILL NOT OVERRIDE EXISTING KEYS. If a key already exists, it will be skipped.

        :param prefix: Filter prefix
        :param dictionary: Dictionary to parse
        :param subcast_keys: The type to cast the keys to
        :param subcast_values:  The type to cast the values to
        :return: AttrDict
        """
        for _key, _value in dictionary.items():
            if _key.startswith(prefix) or _key.startswith(prefix.upper()):
                # Remove the prefix from the key
                clean_key = _key.split('__', 1)[1]
                subkeys = clean_key.split('__')
                _dict = response_dict
                for subkey in subkeys[:-1]:
                    subkey = subcast_keys.deserialize(subkey)
                    if subkey not in _dict:
                        _dict[subkey] = AttrDict()
                    _dict = _dict[subkey]

                final_key = subcast_keys.deserialize(subkeys[-1])
                if final_key in _dict:
                    continue
                _dict[final_key] = subcast_values.deserialize(_value)
        return response_dict

    def dict(self, prefix, subcast_keys: ma.fields.Field = fields.Str,
             subcast_values: ma.fields.Field = fields.Str) -> AttrDict:
        """
        Get the value of a secret as a dictionary.

        # Key format
        DATABASE__default__HOST=localhost
        DATABASE__default__PORT=5432
        DATABASE__default__USER=postgres
        DATABASE__default__PASSWORD=postgres
        DATABASE__default__NAME=postgres
        DATABASE__default__OPTIONS__sslmode=disable

        # Returns
        {
            'default': {
                'HOST': 'localhost',
                'PORT': 5432,
                'USER': 'postgres',
                'PASSWORD': 'postgres',
                'NAME': 'postgres',
                'OPTIONS': {
                    'sslmode': 'disable'
                }
            }
        }
        """
        if isinstance(subcast_keys, type):
            subcast_keys = subcast_keys()
        if isinstance(subcast_values, type):
            subcast_values = subcast_values()

        if not self._secrets:
            self.load()
        filter_prefix = f'{prefix}__'

        response = AttrDict()
        # Load all values that begin with prefix
        for secret in reversed(self._secrets.values()):  # We want to get the most recent secret first
            response.update(self._parse_dict(filter_prefix, secret.SecretValues, response,
                                             subcast_keys, subcast_values))

        # Load all environment variables that begin with prefix
        response.update(self._parse_dict(filter_prefix, self.env, response, subcast_keys, subcast_values))

        return response



"""
AWS Secrets Parser
"""
import base64
import json
import os
from collections import OrderedDict
from typing import Optional

from botocore.client import BaseClient
from botocore.exceptions import ClientError

from .dto import GetValue
from .exceptions import define_error


class SecretParser:
    """
    SecretParser connects to AWS and parses the secret
    """
    SERVICE_NAME: str = 'secretsmanager'

    def __init__(self, default_secret_name: str = None, env=None, **aws_kwargs):
        """
        :param secret_name: The AWS Secrets Manager secret name
        :param env: The environment (default: os.environ)
        :param aws_kwargs: AWS connection kwargs for boto3.client
        """
        _env = env or os.environ
        if hasattr(env, 'dump'):
            # Handle environs.Env object
            _env = os.environ.copy()
            _env.update(env.dump())

        self._env = _env
        if default_secret_name is None:
            default_secret_name = self.env.get('SECRET_NAME', None)

        self.client = None
        self.__client_created = False
        self.default_secret_name = default_secret_name

        self.aws_kwargs = aws_kwargs
        self._secrets = OrderedDict()

    @property
    def env(self):
        """
        our handling of the env is strange because we don't want to
        alter the os.environ if we get a environs.Env object.
        """
        return self._env or os.environ

    def connect(self) -> BaseClient:
        """
        Connect to AWS Secrets Manager to the default aws session
        :return: client
        """
        if self.client:
            return self.client
        import boto3
        self.aws_kwargs.setdefault('region_name', self.env.get('AWS_REGION', 'us-east-1'))

        self.client = boto3.client(self.SERVICE_NAME, **self.aws_kwargs)
        self.__client_created = True
        return self.client

    def load(self, secret_name: str = None, client: BaseClient = None,
             required: bool = False) -> Optional[GetValue]:
        """
        Load secret from AWS Secrets Manager.
        If no secret_name is provided, the default_secret_name is used.
        If required, we will raise an exception if the secret is not found.
        :param secret_name: The AWS Secrets Manager secret name (defaults to default_secret_name)
        :param client: The boto3 client
        :param required: If the secret is required (default: False)
        """
        if secret_name is None:
            secret_name = self.default_secret_name
        if secret_name in self._secrets:
            return self._secrets[secret_name]
        if client is None:
            client = self.client
        if client is None:
            client = self.connect()
        try:
            response = self._load_secret(secret_name, client)
            self._secrets[secret_name] = response
            return response
        except ClientError as error:
            if required:
                raise error
            return None

    def _load_secret(self, secret_name: str, client: BaseClient) -> GetValue:
        try:
            raw_secret = client.get_secret_value(
                SecretId=secret_name
            )
        except ClientError as error:
            error_code = error.response['Error']['Code']
            err = define_error(error_code)
            str(err)
            raise error

        else:
            if raw_secret.get('SecretBinary'):
                raw_secret['SecretValues'] = base64.b64decode(raw_secret['SecretBinary'])
                del raw_secret['SecretBinary']
            else:
                value = raw_secret['SecretString']
                if isinstance(value, str):
                    value = json.loads(value)
                raw_secret['SecretValues'] = value
                del raw_secret['SecretString']

            return GetValue(**raw_secret)

    def close(self):
        try:
            if self.__client_created and self.client:
                self.client.close()
        finally:
            self.client = None
            self._secrets = OrderedDict()

    # Context Management
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()
